<?php

namespace App\Http\Controllers\Api;

use App\Http\Controllers\Controller;
use App\Http\Resources\MenuResource;
use Illuminate\Http\Request;
use App\Models\Menu;

/**
 * @author Arleme Johnson
* Menu Controller Class
*/
class MenuController extends Controller {
    /**
     * @unauthenticated
     * List all menus records and return a collection of menu resources
     * 
     * @param Request $request
     * @return MenuResource::collection
     */
    public function list(Request $request) {
        $request->validate([
            'per_page' => 'nullable|numeric',
            'page' => 'nullable|numeric',
            'type' => 'nullable|in:link,dropdown'
        ]);

        $perPage = $request->per_page ?? 10;
        $type = $request->type ?? null;

        if ($type) {
            $menu = Menu::where('type', $type)->paginate($perPage);
        } else {
            $menu = Menu::paginate($perPage);
        }
        return MenuResource::collection($menu);
    }

    /**
     * Store a new menu record and return the menu resource
     * 
     * @param Request $request
     * @return MenuResource
     */
    public function store(Request $request) {
        $request->validate([
            'name' => 'required',
            'type' => 'required|in:link,dropdown',
            'url' => 'nullable|string',
            'parent_id' => 'nullable|string|exists:menus,id',
            'order' => 'nullable|integer',
            'navbar' => 'required|in:nav-1,nav-2,footer'
        ]);

        $menu = Menu::create($request->all());
        return response(MenuResource::make($menu), 201);
    }

    /**
     * @unauthenticated
     * Get data of a menu record from its id and return the menu resource
     * 
     * @param Request $request
     * @param string $id
     * @return MenuResource
     */
    public function show(Request $request, $id) {
        $menu = Menu::findorFail($id);
        return MenuResource::make($menu);
    }

    /**
     * Update a menu record from its id and return the menu resource
     *
     * @param Request $request
     * @param string $id (uuid format)
     * @return MenuResource
    */
    public function update(Request $request, $id) {
        $menu = Menu::findorFail($id);

        $request->validate([
            'name' => 'required',
            'type' => 'required|in:link,dropdown',
            'url' => 'nullable|string',
            'parent_id' => 'nullable|string|exists:menus,id',
            'order' => 'nullable|integer',
            'navbar' => 'required|in:nav-1,nav-2,footer'
        ]);

        if ($menu) {
            $menu->update($request->all());
            return MenuResource::make($menu);
        }
    }

    /**
     * Delete a menu record from its id
     * 
     * @param Request $request
     * @param string $id (uuid format)
     * @throws ModelNotFoundException if the menu record is not found
     */
    public function delete(Request $request, $id) {
        $menu = Menu::findorFail($id);
        $menu->delete();
        abort(204, 'Menu deleted successfully.'); 
    }
}
