<?php

namespace App\Http\Controllers\Api;

use App\Http\Controllers\Controller;
use App\Http\Resources\PageResource;
use Illuminate\Http\Request;
use App\Models\Page;

/**
 * @author Arleme Johnson
 * Page Controller Class
 */
class PageController extends Controller
{
    /**
     * List all pages records and return a collection of page resources
     * 
     * @param Request $request
     * @return PageResource
     */
    public function list(Request $request)
    {
        $request->validate([
            'per_page' => 'nullable|numeric',
            'page' => 'nullable|numeric',
            // Filtering
            'slug' => 'nullable|string|exists:pages,slug',
            'group_id' => 'nullable|string|exists:groups,id',
            'parent_id' => 'nullable|string|exists:pages,id'
        ]);

        $pages = new Page;

        $pages = $request->group_id ? $pages->where('parent_id', $request->group_id) : $pages;
        $pages = $request->slug ? $pages->where('parent_id', $request->slug) : $pages;
        $pages = $request->parent_id ? $pages->where('parent_id', $request->parent_id) : $pages->where('parent_id', null);

        return response()->json(PageResource::collection($pages->get()));
    }


    /**
     * Get data of a page record from its id and return the page resource
     * 
     * @param Request $request
     * @param string $id (uuid format)
     * @return PageResource
     */
    public function show(Request $request, $id)
    {
        $page = Page::findOrFail($id);
        return PageResource::make($page);
    }

    /**
     * Store a new page record and return the page resource
     * 
     * @param Request $request
     * @return PageResource
     */
    public function store(Request $request)
    {
        $request->validate([
            'title' => 'required|string',
            'slug' => 'required|string|unique:pages,slug',
            'status' => 'required|in:draft,published',
            'parent_id' => 'nullable|string|exists:pages,id',
            'group_id' => 'nullable|string|exists:groups,id',
        ]);

        $page = Page::create($request->all());
        return response(PageResource::make($page), 201);
    }

    /**
     * Update a page record that match an id and return the page resource
     * 
     * @param Request $request
     * @param string $id (uuid format)
     * @return PageResource
     */
    public function update(Request $request, $id)
    {
        // Validate if the id exists
        $page = Page::findOrFail($id);

        $request->validate([
            'title' => 'required|string',
            'slug' => 'required|string|unique:pages,slug',
            'status' => 'required|in:draft,published',
            'parent_id' => 'nullable|string|exists:pages,id',
            'group_id' => 'nullable|string|exists:groups,id'
        ]);
        $page->fill($request->all());
        $page->save();

        return $page;
    }

    /**
     * Delete a page record that match an id
     * 
     * @param Request $request
     * @param string $id (uuid format)
     * @return \Illuminate\Http\Response
     */
    public function delete(Request $request, $id)
    {
        $page = Page::findorFail($id);
        $page->delete();

        abort(204, 'Page deleted successfully.');
    }
}
